@@ -95,9 +95,15 @@ module Tk::BLT::Treeview::ConfigMethod
   end
   private :__item_pathname
 
+  def column_cget_tkstring(name, option)
+    itemcget_tkstring(['column', name], option)
+  end
   def column_cget(name, option)
     itemcget(['column', name], option)
   end
+  def column_cget_strict(name, option)
+    itemcget_strict(['column', name], option)
+  end
   def column_configure(name, slot, value=None)
     itemconfigure(['column', name], slot, value)
   end
@@ -108,9 +114,15 @@ module Tk::BLT::Treeview::ConfigMethod
     current_itemconfiginfo(['column', name], slot)
   end
 
+  def button_cget_tkstring(option)
+    itemcget_tkstring('button', option)
+  end
   def button_cget(option)
     itemcget('button', option)
   end
+  def button_cget_strict(option)
+    itemcget_strict('button', option)
+  end
   def button_configure(slot, value=None)
     itemconfigure('button', slot, value)
   end
@@ -121,6 +133,9 @@ module Tk::BLT::Treeview::ConfigMethod
     current_itemconfiginfo('button', slot)
   end
 
+  def entry_cget_tkstring(option)
+    itemcget_tkstring('entry', option)
+  end
   def entry_cget(option)
     ret = itemcget('entry', option)
     if option == 'bindtags' || option == :bindtags
@@ -129,6 +144,14 @@ module Tk::BLT::Treeview::ConfigMethod
       ret
     end
   end
+  def entry_cget_strict(option)
+    ret = itemcget_strict('entry', option)
+    if option == 'bindtags' || option == :bindtags
+      ret.collect{|tag| TkBindTag.id2obj(tag)}
+    else
+      ret
+    end
+  end
   def entry_configure(slot, value=None)
     itemconfigure('entry', slot, value)
   end
@@ -167,9 +190,15 @@ module Tk::BLT::Treeview::ConfigMethod
     ret
   end
 
+  def sort_cget_tkstring(option)
+    itemcget_tkstring('sort', option)
+  end
   def sort_cget(option)
     itemcget('sort', option)
   end
+  def sort_cget_strict(option)
+    itemcget_strict('sort', option)
+  end
   def sort_configure(slot, value=None)
     itemconfigure('sort', slot, value)
   end
@@ -180,9 +209,15 @@ module Tk::BLT::Treeview::ConfigMethod
     current_itemconfiginfo('sort', slot)
   end
 
+  def text_cget_tkstring(option)
+    itemcget_tkstring('text', option)
+  end
   def text_cget(option)
     itemcget('text', option)
   end
+  def text_cget_strict(option)
+    itemcget_strict('text', option)
+  end
   def text_configure(slot, value=None)
     itemconfigure('text', slot, value)
   end
@@ -193,14 +228,14 @@ module Tk::BLT::Treeview::ConfigMethod
     current_itemconfiginfo('text', slot)
   end
 
-  private :itemcget, :itemconfigure
-  private :itemconfiginfo, :current_itemconfiginfo
+  private :itemcget_tkstring, :itemcget, :itemcget_strict
+  private :itemconfigure, :itemconfiginfo, :current_itemconfiginfo
 end
 
 class Tk::BLT::Treeview
   TkCommandNames = ['::blt::treeview'.freeze].freeze
   WidgetClassName = 'TreeView'.freeze
-  WidgetClassNames[WidgetClassName] = self
+  WidgetClassNames[WidgetClassName] ||= self
 
   include Scrollable
   include ValidateConfigure
@@ -210,7 +245,7 @@ class Tk::BLT::Treeview
   ########################
 
   def __boolval_optkeys
-    ['autocreate', 'allowduplicates', 'exportselection', 'flat', 'hideroot', 
+    ['autocreate', 'allowduplicates', 'exportselection', 'flat', 'hideroot',
       'newtags', 'showtitles', 'sortselection']
   end
   private :__boolval_optkeys
@@ -225,20 +260,38 @@ class Tk::BLT::Treeview
   class OpenCloseCommand < TkValidateCommand
     class ValidateArgs < TkUtil::CallbackSubst
       KEY_TBL = [
-        [ ?W, ?w, :widget ], 
-        [ ?p, ?s, :name ], 
-        [ ?P, ?s, :fullpath ], 
-        [ ?#, ?x, :node_id ], 
+        [ ?W, ?w, :widget ],
+        [ ?p, ?s, :name ],
+        [ ?P, ?s, :fullpath ],
+        [ ?#, ?x, :node_id ],
         nil
       ]
 
       PROC_TBL = [
-        [ ?x, TkComm.method(:num_or_str) ], 
-        [ ?s, TkComm.method(:string) ], 
-        [ ?w, TkComm.method(:window) ], 
+        [ ?x, TkComm.method(:num_or_str) ],
+        [ ?s, TkComm.method(:string) ],
+        [ ?w, TkComm.method(:window) ],
         nil
       ]
 
+=begin
+      # for Ruby m17n :: ?x --> String --> char-code ( getbyte(0) )
+      KEY_TBL.map!{|inf|
+        if inf.kind_of?(Array)
+          inf[0] = inf[0].getbyte(0) if inf[0].kind_of?(String)
+          inf[1] = inf[1].getbyte(0) if inf[1].kind_of?(String)
+        end
+        inf
+      }
+
+      PROC_TBL.map!{|inf|
+        if inf.kind_of?(Array)
+          inf[0] = inf[0].getbyte(0) if inf[0].kind_of?(String)
+        end
+        inf
+      }
+=end
+
       _setup_subst_table(KEY_TBL, PROC_TBL);
 
       def self.ret_val(val)
@@ -273,8 +326,12 @@ class Tk::BLT::Treeview
   ########################
 
   def __destroy_hook__
-    Tk::BLT::Treeview::Node::TreeNodeID_TBL.delete(@path)
-    Tk::BLT::Treeview::Tag::TreeTagID_TBL.delete(@path)
+    Tk::BLT::Treeview::Node::TreeNodeID_TBL.mutex.synchronize{
+      Tk::BLT::Treeview::Node::TreeNodeID_TBL.delete(@path)
+    }
+    Tk::BLT::Treeview::Tag::TreeTagID_TBL.mutex.synchronize{
+      Tk::BLT::Treeview::Tag::TreeTagID_TBL.delete(@path)
+    }
   end
 
   def tagid(tag)
@@ -282,7 +339,7 @@ class Tk::BLT::Treeview
       || tag.kind_of?(Tk::BLT::Treeview::Tag)
       tag.id
     else
-      tag  # maybe an Array of configure paramters
+      tag  # maybe an Array of configure parameters
     end
   end
   private :tagid
@@ -421,7 +478,7 @@ class Tk::BLT::Treeview
     self
   end
   def entry_children(tag, first=None, last=None)
-    simplelist(tk_send('entry', 'children', tagid(tag), 
+    simplelist(tk_send('entry', 'children', tagid(tag),
                        first, last)).collect{|id| tagid2obj(id)}
   end
   def entry_delete(tag, first=None, last=None)
@@ -458,20 +515,38 @@ class Tk::BLT::Treeview
   class FindExecFlagValue < TkValidateCommand
     class ValidateArgs < TkUtil::CallbackSubst
       KEY_TBL = [
-        [ ?W, ?w, :widget ], 
-        [ ?p, ?s, :name ], 
-        [ ?P, ?s, :fullpath ], 
-        [ ?#, ?x, :node_id ], 
+        [ ?W, ?w, :widget ],
+        [ ?p, ?s, :name ],
+        [ ?P, ?s, :fullpath ],
+        [ ?#, ?x, :node_id ],
         nil
       ]
 
       PROC_TBL = [
-        [ ?x, TkComm.method(:num_or_str) ], 
-        [ ?s, TkComm.method(:string) ], 
-        [ ?w, TkComm.method(:window) ], 
+        [ ?x, TkComm.method(:num_or_str) ],
+        [ ?s, TkComm.method(:string) ],
+        [ ?w, TkComm.method(:window) ],
         nil
       ]
 
+=begin
+      # for Ruby m17n :: ?x --> String --> char-code ( getbyte(0) )
+      KEY_TBL.map!{|inf|
+        if inf.kind_of?(Array)
+          inf[0] = inf[0].getbyte(0) if inf[0].kind_of?(String)
+          inf[1] = inf[1].getbyte(0) if inf[1].kind_of?(String)
+        end
+        inf
+      }
+
+      PROC_TBL.map!{|inf|
+        if inf.kind_of?(Array)
+          inf[0] = inf[0].getbyte(0) if inf[0].kind_of?(String)
+        end
+        inf
+      }
+=end
+
       _setup_subst_table(KEY_TBL, PROC_TBL);
 
       def self.ret_val(val)
@@ -487,7 +562,8 @@ class Tk::BLT::Treeview
   def _find_exec_flag_value(val)
     if val.kind_of?(Array)
       cmd, *args = val
-      FindExecFlagValue.new(cmd, args.join(' '))
+      #FindExecFlagValue.new(cmd, args.join(' '))
+      FindExecFlagValue.new(cmd, *args)
     elsif TkComm._callback_entry?(val)
       FindExecFlagValue.new(val)
     else
@@ -585,7 +661,7 @@ class Tk::BLT::Treeview
     }
   end
   def range_open(first, last)
-    simplelist(tk_send('range', '-open', 
+    simplelist(tk_send('range', '-open',
                        tagid(first), tagid(last))).collect{|id|
       tagid2obj(id)
     }
@@ -967,22 +1043,47 @@ class Tk::BLT::Treeview::Node < TkObject
   include Tk::BLT::Treeview::TagOrID_Methods
 
   TreeNodeID_TBL = TkCore::INTERP.create_table
-  TreeNode_ID = ['blt_treeview_node'.freeze, '00000'.taint].freeze
 
-  TkCore::INTERP.init_ip_env{ TreeNodeID_TBL.clear }
+  (TreeNode_ID = ['blt_treeview_node'.freeze, TkUtil.untrust('00000')]).instance_eval{
+    @mutex = Mutex.new
+    def mutex; @mutex; end
+    freeze
+  }
+
+  TkCore::INTERP.init_ip_env{
+    TreeNodeID_TBL.mutex.synchronize{ TreeNodeID_TBL.clear }
+  }
 
   def self.id2obj(tree, id)
     tpath = tree.path
-    return id unless TreeNodeID_TBL[tpath]
-    if TreeNodeID_TBL[tpath][id]
-      TreeNodeID_TBL[tpath][id]
-    else
-      begin
-        self.new(tree, nil, nil, 'node'=>Integer(id))
-      rescue
+    TreeNodeID_TBL.mutex.synchronize{
+      if TreeNodeID_TBL[tpath]
+        if TreeNodeID_TBL[tpath][id]
+          TreeNodeID_TBL[tpath][id]
+        else
+          begin
+            # self.new(tree, nil, nil, 'node'=>Integer(id))
+            unless (tk_call(@tpath, 'get', id)).empty?
+              id = Integer(id)
+              (obj = self.allocate).instance_eval{
+                @parent = @tree = tree
+                @tpath = @parent.path
+                @path = @id = id
+                TreeNodeID_TBL[@tpath] ||= {}
+                TreeNodeID_TBL[@tpath][@id] = self
+              }
+              obj
+            else
+              id
+            end
+          rescue
+            id
+          end
+        end
+      else
         id
       end
-    end
+    }
   end
 
   def self.new(tree, pos, parent=nil, keys={})
@@ -994,13 +1095,21 @@ class Tk::BLT::Treeview::Node < TkObject
     keys = _symbolkey2str(keys)
     tpath = tree.path
 
-    if (id = keys['node']) && (obj = TreeNodeID_TBL[tpath][id])
-      keys.delete('node')
-      tk_call(tree.path, 'move', id, pos, parent) if parent
-      return obj
-    end
+    TreeNodeID_TBL.mutex.synchronize{
+      TreeNodeID_TBL[tpath] ||= {}
+      if (id = keys['node']) && (obj = TreeNodeID_TBL[tpath][id])
+        keys.delete('node')
+        tk_call(tree.path, 'move', id, pos, parent) if parent
+        return obj
+      end
 
-    super(tree, pos, parent, keys)
+      #super(tree, pos, parent, keys)
+      (obj = self.allocate).instance_eval{
+        initialize(tree, pos, parent, keys)
+        TreeNodeID_TBL[tpath][@id] = self
+      }
+      obj
+    }
   end
 
   def initialize(tree, pos, parent, keys)
@@ -1008,16 +1117,23 @@ class Tk::BLT::Treeview::Node < TkObject
     @tpath = @parent.path
 
     if (id = keys['node'])
+      # if tk_call(@tpath, 'get', id).empty?
+      #   fail RuntimeError, "not exist the node '#{id}'"
+      # end
       @path = @id = id
       tk_call(@tpath, 'move', @id, pos, tagid(parent)) if parent
+      configure(keys) if keys && ! keys.empty?
     else
-      name = TreeNode_ID.join(TkCore::INTERP._ip_id_).freeze
-      TreeNode_ID[1].succ!
+      name = nil
+      TreeNode_ID.mutex.synchronize{
+        name = TreeNode_ID.join(TkCore::INTERP._ip_id_).freeze
+        TreeNode_ID[1].succ!
+      }
 
       at = keys.delete['at']
 
       if parent
-        if parent.kind_of?(Tk::BLT::Treeview::Node) || 
+        if parent.kind_of?(Tk::BLT::Treeview::Node) ||
             parent.kind_of?(Tk::BLT::Treeview::Tag)
           path = [get_full(parent.id)[0], name]
           at = nil # ignore 'at' option
@@ -1035,9 +1151,6 @@ class Tk::BLT::Treeview::Node < TkObject
       end
       @path = @id
     end
-
-    TreeNodeID_TBL[@tpath] = {} unless TreeNodeID_TBL[@tpath]
-    TreeNodeID_TBL[@tpath][@id] = self
   end
 
   def id
@@ -1051,37 +1164,66 @@ class Tk::BLT::Treeview::Tag < TkObject
   include Tk::BLT::Treeview::TagOrID_Methods
 
   TreeTagID_TBL = TkCore::INTERP.create_table
-  TreeTag_ID = ['blt_treeview_tag'.freeze, '00000'.taint].freeze
 
-  TkCore::INTERP.init_ip_env{ TreeTagID_TBL.clear }
+  (TreeTag_ID = ['blt_treeview_tag'.freeze, TkUtil.untrust('00000')]).instance_eval{
+    @mutex = Mutex.new
+    def mutex; @mutex; end
+    freeze
+  }
 
-  def self.id2obj(tree, id)
+  TkCore::INTERP.init_ip_env{
+    TreeTagID_TBL.mutex.synchronize{ TreeTagID_TBL.clear }
+  }
+
+  def self.id2obj(tree, name)
     tpath = tree.path
-    return id unless TreeTagID_TBL[tpath]
-    if TreeTagID_TBL[tpath][id]
-      TreeTagID_TBL[tpath][id]
-    else
-      begin
-        self.new(tree, nil, nil, 'name'=>Integer(id))
-      rescue
+    TreeTagID_TBL.mutex.synchronize{
+      if TreeTagID_TBL[tpath]
+        if TreeTagID_TBL[tpath][name]
+          TreeTagID_TBL[tpath][name]
+        else
+          #self.new(tree, name)
+          (obj = self.allocate).instance_eval{
+            @parent = @tree = tree
+            @tpath = @parent.path
+            @path = @id = name
+            TreeTagID_TBL[@tpath] = {} unless TreeTagID_TBL[@tpath]
+            TreeTagID_TBL[@tpath][@id] = self
+          }
+          obj
+        end
+      else
         id
       end
-    end
+    }
   end
 
   def self.new_by_name(tree, name, *ids)
-    if (obj = TreeTagID_TBL[tree.path][name])
-      return obj
-    end
-    new([tree, name], ids)
+    TreeTagID_TBL.mutex.synchronize{
+      unless (obj = TreeTagID_TBL[tree.path][name])
+        (obj = self.allocate).instance_eval{
+          initialize(tree, name, ids)
+          TreeTagID_TBL[@tpath] = {} unless TreeTagID_TBL[@tpath]
+          TreeTagID_TBL[@tpath][@id] = self
+        }
+      end
+      obj
+    }
   end
 
   def self.new(tree, *ids)
-    if tree.kind_of?(Array)
-      super(tree[0], tree[1], ids)
-    else
-      super(tree, nil, ids)
-    end
+    TreeTagID_TBL.mutex.synchronize{
+      (obj = self.allocate).instance_eval{
+        if tree.kind_of?(Array)
+          initialize(tree[0], tree[1], ids)
+        else
+          initialize(tree, nil, ids)
+        end
+        TreeTagID_TBL[@tpath] = {} unless TreeTagID_TBL[@tpath]
+        TreeTagID_TBL[@tpath][@id] = self
+      }
+      obj
+    }
   end
 
   def initialize(tree, name, ids)
@@ -1091,13 +1233,12 @@ class Tk::BLT::Treeview::Tag < TkObject
     if name
       @path = @id = name
     else
-      @path = @id = TreeTag_ID.join(TkCore::INTERP._ip_id_).freeze
-      TreeTag_ID[1].succ!
+      TreeTag_ID.mutex.synchronize{
+        @path = @id = TreeTag_ID.join(TkCore::INTERP._ip_id_).freeze
+        TreeTag_ID[1].succ!
+      }
     end
 
-    TreeTagID_TBL[@tpath] = {} unless TreeTagID_TBL[@tpath]
-    TreeTagID_TBL[@tpath][@id] = self
-
     unless ids.empty?
       tk_call(@tpath, 'tag', 'add', @id, *(ids.collect{|id| tagid(id)}))
     end
@@ -1142,5 +1283,5 @@ end
 class Tk::BLT::Hiertable
   TkCommandNames = ['::blt::hiertable'.freeze].freeze
   WidgetClassName = 'Hiertable'.freeze
-  WidgetClassNames[WidgetClassName] = self
+  WidgetClassNames[WidgetClassName] ||= self
 end
