@@ -1,5 +1,4 @@
 require 'test/unit'
-require_relative 'envutil'
 
 class TestSetTraceFunc < Test::Unit::TestCase
   def setup
@@ -285,11 +284,9 @@ class TestSetTraceFunc < Test::Unit::TestCase
 
     [["c-return", 1, :set_trace_func, Kernel],
      ["line", 4, __method__, self.class],
-     ["c-call", 4, :any?, Enumerable],
-     ["c-call", 4, :each, Array],
+     ["c-call", 4, :any?, Array],
      ["line", 4, __method__, self.class],
-     ["c-return", 4, :each, Array],
-     ["c-return", 4, :any?, Enumerable],
+     ["c-return", 4, :any?, Array],
      ["line", 5, __method__, self.class],
      ["c-call", 5, :set_trace_func, Kernel]].each{|e|
       assert_equal(e, events.shift)
@@ -509,7 +506,6 @@ class TestSetTraceFunc < Test::Unit::TestCase
      [:return,  16, "xyzzy", xyzzy.class, :bar,             xyzzy,       :XYZZY_bar, xyzzy],
      [:return,  12, "xyzzy", xyzzy.class, :foo,             xyzzy,       :XYZZY_foo, xyzzy],
      [:line,    20, "xyzzy", TestSetTraceFunc, method,      self,        :outer, :nothing],
-     [:line,    20, "xyzzy", TestSetTraceFunc, method,      self,        :outer, :nothing],
      [:c_call,  20, "xyzzy", Kernel,      :raise,           self,        :outer, :nothing],
      [:c_call,  20, "xyzzy", Exception,   :exception,       RuntimeError, :outer, :nothing],
      [:c_call,  20, "xyzzy", Exception,   :initialize,      raised_exc,  :outer, :nothing],
@@ -565,14 +561,18 @@ class TestSetTraceFunc < Test::Unit::TestCase
   def test_tracepoint
     events1, answer_events = *trace_by_tracepoint(:line, :class, :end, :call, :return, :c_call, :c_return, :raise)
 
-    mesg = events1.map{|e|
-      if false
-        p [:event, e[0]]
-        p [:line_file, e[1], e[2]]
-        p [:id, e[4]]
+    ms = [events1, answer_events].map{|evs|
+      evs.map{|e|
+        "#{e[0]} - #{e[2]}:#{e[1]} id: #{e[4]}"
+      }
+    }
+
+    mesg = ms[0].zip(ms[1]).map{|a, b|
+      if a != b
+        "#{a} <-> #{b}"
       end
-      "#{e[0]} - #{e[2]}:#{e[1]} id: #{e[4]}"
-    }.join("\n")
+    }.compact.join("\n")
+
     answer_events.zip(events1){|answer, event|
       assert_equal answer, event, mesg
     }
@@ -977,6 +977,27 @@ class TestSetTraceFunc < Test::Unit::TestCase
     end
   end
 
+  def test_trace_point_binding_after_break
+    bug10689 = '[ruby-dev:48797]'
+    assert_in_out_err([], <<-INPUT, [], [], bug10689)
+      class Bug
+        include Enumerable
+
+        def each
+          [0].each do
+            yield
+          end
+        end
+      end
+
+      TracePoint.trace(:c_return) do |tp|
+        tp.binding
+      end
+
+      Bug.new.all? { false }
+    INPUT
+  end
+
   def test_tracepoint_b_return_with_next
     n = 0
     TracePoint.new(:b_return){
@@ -1028,6 +1049,7 @@ class TestSetTraceFunc < Test::Unit::TestCase
   def test_a_call
     events = []
     TracePoint.new(:a_call){|tp|
+      next if !target_thread?
       events << tp.event
     }.enable{
       1.times{
@@ -1049,6 +1071,7 @@ class TestSetTraceFunc < Test::Unit::TestCase
   def test_a_return
     events = []
     TracePoint.new(:a_return){|tp|
+      next if !target_thread?
       events << tp.event
     }.enable{
       1.times{
@@ -1072,6 +1095,7 @@ class TestSetTraceFunc < Test::Unit::TestCase
     events = []
     assert !defined?(MISSING_CONSTANT_59398)
     TracePoint.new(:c_call, :c_return, :call, :return){|tp|
+      next if !target_thread?
       next unless tp.defined_class == Module
       # rake/ext/module.rb aliases :const_missing and Ruby uses the aliased name
       # but this only happens when running the full test suite
@@ -1102,6 +1126,7 @@ class TestSetTraceFunc < Test::Unit::TestCase
     events = []
     aliased = AliasedRubyMethod.new
     TracePoint.new(:call, :return){|tp|
+      next if !target_thread?
       events << [tp.event, tp.method_id]
     }.enable{
       aliased.bar
@@ -1120,6 +1145,7 @@ class TestSetTraceFunc < Test::Unit::TestCase
     events = []
     aliased = AliasedCMethod.new
     TracePoint.new(:call, :return, :c_call, :c_return){|tp|
+      next if !target_thread?
       events << [tp.event, tp.method_id]
     }.enable{
       aliased.size
@@ -1137,6 +1163,7 @@ class TestSetTraceFunc < Test::Unit::TestCase
     events = []
     assert !respond_to?(:missing_method_59398)
     TracePoint.new(:c_call, :c_return, :call, :return){|tp|
+      next if !target_thread?
       next unless tp.defined_class == BasicObject
       # rake/ext/module.rb aliases :const_missing and Ruby uses the aliased name
       # but this only happens when running the full test suite
@@ -1197,61 +1224,6 @@ class TestSetTraceFunc < Test::Unit::TestCase
     }, '[Bug #9940]'
   end
 
-  def method_test_rescue_should_not_cause_b_return
-    begin
-      raise
-    rescue
-      return
-    end
-  end
-
-  def method_test_ensure_should_not_cause_b_return
-    begin
-      raise
-    ensure
-      return
-    end
-  end
-
-  def test_rescue_and_ensure_should_not_cause_b_return
-    curr_thread = Thread.current
-    trace = TracePoint.new(:b_call, :b_return){
-      next if curr_thread != Thread.current
-      flunk("Should not reach here because there is no block.")
-    }
-
-    begin
-      trace.enable
-      method_test_rescue_should_not_cause_b_return
-      begin
-        method_test_ensure_should_not_cause_b_return
-      rescue
-        # ignore
-      end
-    ensure
-      trace.disable
-    end
-  end
-
-  define_method(:method_test_argument_error_on_bmethod){|correct_key: 1|}
-
-  def test_argument_error_on_bmethod
-    events = []
-    curr_thread = Thread.current
-    TracePoint.new(:call, :return){|tp|
-      next if curr_thread != Thread.current
-      events << [tp.event, tp.method_id]
-    }.enable do
-      begin
-        method_test_argument_error_on_bmethod(wrong_key: 2)
-      rescue => e
-        # ignore
-      end
-    end
-
-    assert_equal [], events # should be empty.
-  end
-
   def method_prefix event
     case event
     when :call, :return
@@ -1269,78 +1241,76 @@ class TestSetTraceFunc < Test::Unit::TestCase
 
   def assert_consistent_call_return message='', check_events: nil
     check_events ||= %i(a_call a_return)
-    call_events = []
-    return_events = []
+    call_stack = []
 
     TracePoint.new(*check_events){|tp|
       next unless target_thread?
 
       case tp.event.to_s
       when /call/
-        call_events << method_label(tp)
+        call_stack << method_label(tp)
       when /return/
-        return_events << method_label(tp)
+        frame = call_stack.pop
+        assert_equal(frame, method_label(tp))
       end
     }.enable do
       yield
     end
 
-    assert_equal false, call_events.empty?
-    assert_equal false, return_events.empty?
-    assert_equal call_events, return_events.reverse, message
+    assert_equal true, call_stack.empty?
+  end
+
+  def method_test_rescue_should_not_cause_b_return
+    begin
+      raise
+    rescue
+      return
+    end
+  end
+
+  def method_test_ensure_should_not_cause_b_return
+    begin
+      raise
+    ensure
+      return
+    end
+  end
+
+  def test_rescue_and_ensure_should_not_cause_b_return
+    assert_consistent_call_return '[Bug #9957]' do
+      method_test_rescue_should_not_cause_b_return
+      begin
+        method_test_ensure_should_not_cause_b_return
+      rescue
+        # ignore
+      end
+    end
+  end
+
+  define_method(:method_test_argument_error_on_bmethod){|correct_key: 1|}
+
+  def test_argument_error_on_bmethod
+    assert_consistent_call_return '[Bug #9959]' do
+      begin
+        method_test_argument_error_on_bmethod(wrong_key: 2)
+      rescue => e
+        # ignore
+      end
+    end
   end
 
   def test_rb_rescue
-    events = []
-    curr_thread = Thread.current
-    TracePoint.new(:a_call, :a_return){|tp|
-      next if curr_thread != Thread.current
-      events << [tp.event, tp.method_id]
-    }.enable do
+    assert_consistent_call_return '[Bug #9961]' do
       begin
         -Numeric.new
       rescue => e
         # ignore
       end
     end
-
-    assert_equal [
-    [:b_call, :test_rb_rescue],
-      [:c_call, :new],
-        [:c_call, :initialize],
-        [:c_return, :initialize],
-      [:c_return, :new],
-      [:c_call, :-@],
-        [:c_call, :coerce],
-          [:c_call, :new],
-            [:c_call, :initialize],
-            [:c_return, :initialize],
-          [:c_return, :new],
-          [:c_call, :exception],
-          [:c_return, :exception],
-          [:c_call, :backtrace],
-          [:c_return, :backtrace],
-        [:c_return, :coerce],            # don't miss it!
-        [:c_call, :to_s],
-        [:c_return, :to_s],
-        [:c_call, :to_s],
-        [:c_return, :to_s],
-        [:c_call, :new],
-          [:c_call, :initialize],
-          [:c_return, :initialize],
-        [:c_return, :new],
-        [:c_call, :exception],
-        [:c_return, :exception],
-        [:c_call, :backtrace],
-        [:c_return, :backtrace],
-      [:c_return, :-@],
-      [:c_call, :===],
-      [:c_return, :===],
-    [:b_return, :test_rb_rescue]], events
   end
 
   def test_b_call_with_redo
-    assert_consistent_call_return do
+    assert_consistent_call_return '[Bug #9964]' do
       i = 0
       1.times{
         break if (i+=1) > 10
@@ -1349,23 +1319,17 @@ class TestSetTraceFunc < Test::Unit::TestCase
     end
   end
 
-  class Bug10724
-    def initialize
-      loop{return}
-    end
-  end
+  def test_no_duplicate_line_events
+    lines = []
+    dummy = []
 
-  def test_throwing_return_with_finish_frame
-    target_th = Thread.current
-    evs = []
-
-    TracePoint.new(:call, :return){|tp|
-      return if Thread.current != target_th
-      evs << tp.event
+    TracePoint.new(:line){|tp|
+      next unless target_thread?
+      lines << tp.lineno
     }.enable{
-      a = Bug10724.new
+      dummy << (1) + (2)
+      dummy << (1) + (2)
     }
-
-    assert_equal([:call, :return], evs)
+    assert_equal [__LINE__ - 3, __LINE__ - 2], lines, 'Bug #10449'
   end
 end
